package sun.security.provider;

import java.security.interfaces.DSAParams;
import java.security.InvalidKeyException;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.ProviderException;
import sun.security.jca.JCAUtil;
import java.security.KeyPair;
import sun.security.util.SecurityProviderConstants;
import java.security.InvalidAlgorithmParameterException;
import java.security.spec.AlgorithmParameterSpec;
import java.security.InvalidParameterException;
import java.security.SecureRandom;
import java.security.spec.DSAParameterSpec;
import java.security.KeyPairGenerator;

class DSAKeyPairGenerator extends KeyPairGenerator
{
    private int plen;
    private int qlen;
    boolean forceNewParameters;
    private DSAParameterSpec params;
    private SecureRandom random;
    
    DSAKeyPairGenerator(int n) {
        super("DSA");
        this.initialize(n, null);
    }
    
    private static void checkStrength(int i, int j) {
        if (i < 512 || i > 1024 || i % 64 != 0 || j != 160) {
            if (i == 2048) {
                if (j != 224) {
                    if (j == 256) {
                        return;
                    }
                }
                else {
                    return;
                }
            }
            if (i != 3072 || j != 256) {
                throw new InvalidParameterException("Unsupported prime and subprime size combination: " + i + ", " + j);
            }
        }
    }
    
    @Override
    public void initialize(int n, SecureRandom secureRandom) {
        this.init(n, secureRandom, false);
    }
    
    @Override
    public void initialize(AlgorithmParameterSpec algorithmParameterSpec, SecureRandom secureRandom) throws InvalidAlgorithmParameterException {
        if (!(algorithmParameterSpec instanceof DSAParameterSpec)) {
            throw new InvalidAlgorithmParameterException("Inappropriate parameter");
        }
        this.init((DSAParameterSpec)algorithmParameterSpec, secureRandom, false);
        return;
    }
    
    void init(int n, SecureRandom random, boolean forceNewParameters) {
        int defDSASubprimeSize = SecurityProviderConstants.getDefDSASubprimeSize(n);
        checkStrength(n, defDSASubprimeSize);
        this.plen = n;
        this.qlen = defDSASubprimeSize;
        this.params = null;
        this.random = random;
        this.forceNewParameters = forceNewParameters;
    }
    
    void init(DSAParameterSpec params, SecureRandom random, boolean forceNewParameters) {
        int bitLength = params.getP().bitLength();
        int bitLength2 = params.getQ().bitLength();
        checkStrength(bitLength, bitLength2);
        this.plen = bitLength;
        this.qlen = bitLength2;
        this.params = params;
        this.random = random;
        this.forceNewParameters = forceNewParameters;
    }
    
    @Override
    public KeyPair generateKeyPair() {
        if (this.random == null) {
            this.random = JCAUtil.getSecureRandom();
        }
        DSAParameterSpec dsaParameterSpec;
        try {
            if (this.forceNewParameters) {
                dsaParameterSpec = ParameterCache.getNewDSAParameterSpec(this.plen, this.qlen, this.random);
            }
            else {
                if (this.params == null) {
                    this.params = ParameterCache.getDSAParameterSpec(this.plen, this.qlen, this.random);
                }
                dsaParameterSpec = this.params;
            }
        }
        catch (GeneralSecurityException cause) {
            throw new ProviderException(cause);
        }
        return this.generateKeyPair(dsaParameterSpec.getP(), dsaParameterSpec.getQ(), dsaParameterSpec.getG(), this.random);
    }
    
    private KeyPair generateKeyPair(BigInteger bigInteger, BigInteger bigInteger2, BigInteger bigInteger3, SecureRandom secureRandom) {
        BigInteger generateX = this.generateX(secureRandom, bigInteger2);
        BigInteger generateY = this.generateY(generateX, bigInteger, bigInteger3);
        try {
            DSAPublicKey publicKey;
            if (DSAKeyFactory.SERIAL_INTEROP) {
                publicKey = new DSAPublicKey(generateY, bigInteger, bigInteger2, bigInteger3);
            }
            else {
                publicKey = new DSAPublicKeyImpl(generateY, bigInteger, bigInteger2, bigInteger3);
            }
            return new KeyPair(publicKey, new DSAPrivateKey(generateX, bigInteger, bigInteger2, bigInteger3));
        }
        catch (InvalidKeyException cause) {
            throw new ProviderException(cause);
        }
    }
    
    private BigInteger generateX(SecureRandom secureRandom, BigInteger bigInteger) {
        byte[] array = new byte[this.qlen];
        BigInteger mod;
        do {
            secureRandom.nextBytes(array);
            mod = new BigInteger(1, array).mod(bigInteger);
        } while (mod.signum() <= 0 || mod.compareTo(bigInteger) >= 0);
        return mod;
    }
    
    BigInteger generateY(BigInteger exponent, BigInteger m, BigInteger bigInteger) {
        return bigInteger.modPow(exponent, m);
    }
    
    public static final class Current extends DSAKeyPairGenerator
    {
        public Current() {
            super(SecurityProviderConstants.DEF_DSA_KEY_SIZE);
        }
    }
    
    public static final class Legacy extends DSAKeyPairGenerator implements DSAKeyPairGenerator
    {
        public Legacy() {
            super(1024);
        }
        
        @Override
        public void initialize(int primeSize, boolean b, SecureRandom secureRandom) throws InvalidParameterException {
            if (b) {
                super.init(primeSize, secureRandom, true);
            }
            else {
                DSAParameterSpec cachedDSAParameterSpec = ParameterCache.getCachedDSAParameterSpec(primeSize, SecurityProviderConstants.getDefDSASubprimeSize(primeSize));
                if (cachedDSAParameterSpec == null) {
                    throw new InvalidParameterException("No precomputed parameters for requested modulus size available");
                }
                super.init(cachedDSAParameterSpec, secureRandom, false);
            }
        }
        
        @Override
        public void initialize(DSAParams dsaParams, SecureRandom secureRandom) throws InvalidParameterException {
            if (dsaParams == null) {
                throw new InvalidParameterException("Params must not be null");
            }
            super.init(new DSAParameterSpec(dsaParams.getP(), dsaParams.getQ(), dsaParams.getG()), secureRandom, false);
            return;
        }
    }
}
